@@ -64,8 +64,8 @@ static void     gs_fade_finalize   (GObject        *object);
 
 struct GSFadePrivate
 {
-        gboolean         enabled;
-        gboolean         active;
+        guint            enabled : 1;
+        guint            active : 1;
 
         guint            timeout;
 
@@ -73,6 +73,9 @@ struct GSFadePrivate
         guint            num_steps;
         guint            timer_id;
 
+        gdouble          alpha_per_iter;
+        gdouble          current_alpha;
+
         int              fade_type;
 
         int              num_screens;
@@ -83,22 +86,17 @@ struct GSFadePrivate
 
 };
 
-enum { 
+enum {
         FADED,
         LAST_SIGNAL
 };
 
-enum {
-        PROP_0
-};
-
 enum {
         FADE_TYPE_NONE,
         FADE_TYPE_GAMMA_NUMBER,
         FADE_TYPE_GAMMA_RAMP
 };
 
-static GObjectClass *parent_class = NULL;
 static guint         signals [LAST_SIGNAL] = { 0, };
 
 G_DEFINE_TYPE (GSFade, gs_fade, G_TYPE_OBJECT)
@@ -189,9 +187,9 @@ xf86_whack_gamma (int              screen,
                 unsigned short *r, *g, *b;
                 int i;
 
-                r = (unsigned short *) malloc (info->size * sizeof (unsigned short));
-                g = (unsigned short *) malloc (info->size * sizeof (unsigned short));
-                b = (unsigned short *) malloc (info->size * sizeof (unsigned short));
+                r = g_new0 (unsigned short, info->size);
+                g = g_new0 (unsigned short, info->size);
+                b = g_new0 (unsigned short, info->size);
 
                 for (i = 0; i < info->size; i++) {
                         r[i] = info->r[i] * ratio;
@@ -245,13 +243,13 @@ check_gamma_extension (void)
                 return FADE_TYPE_NONE;  /* unable to get version number? */
         }
 
-        if (major < XF86_VIDMODE_GAMMA_MIN_MAJOR || 
+        if (major < XF86_VIDMODE_GAMMA_MIN_MAJOR ||
             (major == XF86_VIDMODE_GAMMA_MIN_MAJOR &&
              minor < XF86_VIDMODE_GAMMA_MIN_MINOR)) {
                 return FADE_TYPE_NONE;  /* extension is too old for gamma. */
         }
 
-        if (major < XF86_VIDMODE_GAMMA_RAMP_MIN_MAJOR || 
+        if (major < XF86_VIDMODE_GAMMA_RAMP_MIN_MAJOR ||
             (major == XF86_VIDMODE_GAMMA_RAMP_MIN_MAJOR &&
              minor < XF86_VIDMODE_GAMMA_RAMP_MIN_MINOR)) {
                 return FADE_TYPE_GAMMA_NUMBER;  /* extension is too old for gamma ramps. */
@@ -300,7 +298,7 @@ gamma_info_init (GSFade *fade)
 
         info = g_new0 (xf86_gamma_info, fade->priv->num_screens);
         fade->priv->gamma_info = info;
-        
+
         /* Get the current gamma maps for all screens.
            Bug out and return -1 if we can't get them for some screen.
         */
@@ -330,12 +328,9 @@ gamma_info_init (GSFade *fade)
                                 goto FAIL;
                         }
 
-                        info [screen].r = (unsigned short *)
-                                calloc (info[screen].size, sizeof (unsigned short));
-                        info [screen].g = (unsigned short *)
-                                calloc (info[screen].size, sizeof (unsigned short));
-                        info [screen].b = (unsigned short *)
-                                calloc (info[screen].size, sizeof (unsigned short));
+                        info [screen].r = g_new0 (unsigned short, info[screen].size);
+                        info [screen].g = g_new0 (unsigned short, info[screen].size);
+                        info [screen].b = g_new0 (unsigned short, info[screen].size);
 
                         if (! (info [screen].r && info [screen].g && info [screen].b)) {
                                 goto FAIL;
@@ -438,37 +433,57 @@ gs_fade_set_alpha (GSFade *fade,
 }
 
 static gboolean
-gs_fade_iter (GSFade *fade)
+gs_fade_out_iter (GSFade *fade)
 {
         gboolean ret;
-        gdouble  alpha;
 
-        fade->priv->step++;
-
-        if (fade->priv->step > fade->priv->num_steps) {
+        if (fade->priv->current_alpha < 0.01) {
                 return FALSE;
         }
 
-        alpha = 1.0 - (gdouble)fade->priv->step / (gdouble)fade->priv->num_steps;
+        fade->priv->current_alpha -= fade->priv->alpha_per_iter;
 
-        ret = gs_fade_set_alpha (fade, alpha);
+        ret = gs_fade_set_alpha (fade, fade->priv->current_alpha);
 
         return ret;
 }
 
-static void
+static gboolean
+gs_fade_stop (GSFade *fade)
+{
+        if (fade->priv->timer_id > 0) {
+                g_source_remove (fade->priv->timer_id);
+                fade->priv->timer_id = 0;
+        }
+
+        fade->priv->step = 0;
+        fade->priv->active = FALSE;
+
+        return TRUE;
+}
+
+void
 gs_fade_finish (GSFade *fade)
 {
+        g_return_if_fail (GS_IS_FADE (fade));
+
+        if (! fade->priv->active) {
+                return;
+        }
+
+        gs_fade_stop (fade);
+
         g_signal_emit (fade, signals [FADED], 0);
-        gs_fade_set_active (fade, FALSE);
+
+        fade->priv->active = FALSE;
 }
 
 static gboolean
-fade_timer (GSFade *fade)
+fade_out_timer (GSFade *fade)
 {
         gboolean res;
 
-        res = gs_fade_iter (fade);
+        res = gs_fade_out_iter (fade);
 
         /* if failed then fade is complete */
         if (! res) {
@@ -487,66 +502,7 @@ gs_fade_get_active (GSFade *fade)
         return fade->priv->active;
 }
 
-static gboolean
-gs_fade_start (GSFade *fade)
-{
-        guint steps_per_sec = 30;
-        guint msecs_per_step;
-
-        if (fade->priv->timer_id > 0) {
-                g_warning ("Fade is already active");
-                return FALSE;
-        }
-
-        fade->priv->step = 0;
-        fade->priv->num_steps = (fade->priv->timeout / 1000) * steps_per_sec;
-
-        msecs_per_step = 1000 / steps_per_sec;
-
-        fade->priv->timer_id = g_timeout_add (msecs_per_step, (GSourceFunc)fade_timer, fade);
-
-        return TRUE;
-}
-
-static gboolean
-gs_fade_stop (GSFade *fade)
-{
-        if (fade->priv->timer_id > 0) {
-                g_source_remove (fade->priv->timer_id);
-                fade->priv->timer_id = 0;
-        }
-
-        fade->priv->step = 0;
-
-        return TRUE;
-}
-
-gboolean
-gs_fade_set_active (GSFade  *fade,
-                    gboolean active)
-{
-        g_return_val_if_fail (GS_IS_FADE (fade), FALSE);
-
-        gs_debug ("Setting fade: %s", (active ? "ON" : "OFF"));
-
-        if (fade->priv->fade_type == FADE_TYPE_NONE) {
-                return FALSE;
-        }
-
-        if (fade->priv->active != active) {
-                fade->priv->active = active;
-
-                if (active) {
-                        gs_fade_start (fade);
-                } else {
-                        gs_fade_stop (fade);
-                }
-        }
-
-        return TRUE;
-}
-
-void
+static void
 gs_fade_set_timeout (GSFade   *fade,
                      guint     timeout)
 {
@@ -556,35 +512,113 @@ gs_fade_set_timeout (GSFade   *fade,
 }
 
 static void
-gs_fade_now_faded_callback (GSFade *fade,
-                            int    *flag)
+gs_fade_start (GSFade *fade,
+               guint   timeout)
+{
+        guint steps_per_sec = 30;
+        guint msecs_per_step;
+
+        g_return_if_fail (GS_IS_FADE (fade));
+
+        if (fade->priv->timer_id > 0) {
+                gs_fade_stop (fade);
+        }
+
+        fade->priv->active = TRUE;
+
+        gs_fade_set_timeout (fade, timeout);
+
+        if (fade->priv->fade_type != FADE_TYPE_NONE) {
+                guint num_steps;
+
+                num_steps = (fade->priv->timeout / 1000) * steps_per_sec;
+                msecs_per_step = 1000 / steps_per_sec;
+                fade->priv->alpha_per_iter = 1.0 / (gdouble)num_steps;
+
+                fade->priv->timer_id = g_timeout_add (msecs_per_step, (GSourceFunc)fade_out_timer, fade);
+        } else {
+                gs_fade_finish (fade);
+        }
+}
+
+typedef struct
+{
+        GSFadeDoneFunc done_cb;
+        gpointer       data;
+} FadedCallbackData;
+
+static void
+gs_fade_async_callback (GSFade            *fade,
+                        FadedCallbackData *cdata)
+{
+        g_signal_handlers_disconnect_by_func (fade,
+                                              gs_fade_async_callback, 
+                                              cdata);
+
+        if (cdata->done_cb) {
+                cdata->done_cb (fade, cdata->data);
+        }
+
+        g_free (cdata);
+}
+
+void
+gs_fade_async (GSFade        *fade,
+               guint          timeout,
+               GSFadeDoneFunc func,
+               gpointer       data)
+{
+        g_return_if_fail (GS_IS_FADE (fade));
+
+        /* if fade is active then pause it */
+        if (fade->priv->active) {
+                gs_fade_stop (fade);
+        }
+
+        if (func) {
+                FadedCallbackData *cb_data;
+
+                cb_data = g_new0 (FadedCallbackData, 1);
+                cb_data->done_cb = func;
+                cb_data->data = data;
+
+                g_signal_connect (fade, "faded",
+                                  G_CALLBACK (gs_fade_async_callback),
+                                  cb_data);
+        }
+
+        gs_fade_start (fade, timeout);
+}
+
+static void
+gs_fade_sync_callback (GSFade *fade,
+                       int    *flag)
 {
         *flag = TRUE;
         g_signal_handlers_disconnect_by_func (fade,
-                                              gs_fade_now_faded_callback, 
+                                              gs_fade_sync_callback,
                                               flag);
 }
 
 void
-gs_fade_now (GSFade *fade)
+gs_fade_sync (GSFade        *fade,
+              guint          timeout)
 {
         int      flag = FALSE;
-        gboolean res;
 
         g_return_if_fail (GS_IS_FADE (fade));
 
-        if (fade->priv->fade_type == FADE_TYPE_NONE)
-                return;
-
-        res = gs_fade_set_active (fade, TRUE);
-        if (! res) {
-                return;
+        /* if fade is active then pause it */
+        if (fade->priv->active) {
+                gs_fade_stop (fade);
         }
 
         g_signal_connect (fade, "faded",
-                          G_CALLBACK (gs_fade_now_faded_callback),
+                          G_CALLBACK (gs_fade_sync_callback),
                           &flag);
 
+        gs_fade_start (fade, timeout);
+
         while (! flag) {
                 gtk_main_iteration ();
         }
@@ -593,37 +627,17 @@ gs_fade_now (GSFade *fade)
 void
 gs_fade_reset (GSFade *fade)
 {
+        g_return_if_fail (GS_IS_FADE (fade));
+
+        gs_debug ("Resetting fade");
+
         if (fade->priv->active) {
                 gs_fade_stop (fade);
         }
 
-        gs_fade_set_alpha (fade, 1.0);
-}
+        fade->priv->current_alpha = 1.0;
 
-static void
-gs_fade_set_property (GObject            *object,
-                      guint               prop_id,
-                      const GValue       *value,
-                      GParamSpec         *pspec)
-{
-        switch (prop_id) {
-        default:
-                G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-                break;
-        }
-}
-
-static void
-gs_fade_get_property (GObject            *object,
-                      guint               prop_id,
-                      GValue             *value,
-                      GParamSpec         *pspec)
-{
-        switch (prop_id) {
-        default:
-                G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-                break;
-        }
+        gs_fade_set_alpha (fade, fade->priv->current_alpha);
 }
 
 static void
@@ -631,11 +645,7 @@ gs_fade_class_init (GSFadeClass *klass)
 {
         GObjectClass   *object_class = G_OBJECT_CLASS (klass);
 
-        parent_class = g_type_class_peek_parent (klass);
-
-        object_class->finalize     = gs_fade_finalize;
-        object_class->get_property = gs_fade_get_property;
-        object_class->set_property = gs_fade_set_property;
+        object_class->finalize = gs_fade_finalize;
 
         signals [FADED] =
                 g_signal_new ("faded",
@@ -659,6 +669,7 @@ gs_fade_init (GSFade *fade)
         fade->priv = GS_FADE_GET_PRIVATE (fade);
 
         fade->priv->timeout = 1000;
+        fade->priv->current_alpha = 1.0;
 
         fade->priv->fade_type = check_gamma_extension ();
 
@@ -682,7 +693,7 @@ gs_fade_finalize (GObject *object)
 
         gamma_info_free (fade);
 
-        G_OBJECT_CLASS (parent_class)->finalize (object);
+        G_OBJECT_CLASS (gs_fade_parent_class)->finalize (object);
 }
 
 GSFade *
